==> orquestra-vqa-0.3.0.tar.gz: orquestra-vqa-0.3.0/tests/orquestra/vqa/ansatz/qaoa_warm_start_test.py <==
################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
import numpy as np
import pytest
import sympy
from orquestra.quantum.circuits import RY, RZ, Circuit
from orquestra.quantum.openfermion import (
    IsingOperator,
    QubitOperator,
    change_operator_type,
)
from orquestra.quantum.utils import compare_unitary

from orquestra.vqa.ansatz.qaoa_warm_start import (
    WarmStartQAOAAnsatz,
    convert_relaxed_solution_to_angles,
)
from orquestra.vqa.api.ansatz_test import AnsatzTests


def create_betas(number_of_layers):
    return sympy.symbols(f"beta_:{number_of_layers}")


def create_gammas(number_of_layers):
    return sympy.symbols(f"gamma_:{number_of_layers}")


def create_symbols_map(number_of_layers):
    betas = create_betas(number_of_layers)
    gammas = create_gammas(number_of_layers)
    symbols_map = []
    beta_value = 0.5
    gamma_value = 0.7
    for beta, gamma in zip(betas, gammas):
        symbols_map.append([beta, beta_value])
        symbols_map.append([gamma, gamma_value])
    return symbols_map


def create_target_unitary(thetas, number_of_layers):
    target_circuit = Circuit()
    target_circuit += RY(thetas[0])(0)
    target_circuit += RY(thetas[1])(1)
    betas = create_betas(number_of_layers)
    gammas = create_gammas(number_of_layers)
    symbols_map = create_symbols_map(number_of_layers)
    for layer_id in range(number_of_layers):
        beta = betas[layer_id]
        gamma = gammas[layer_id]
        target_circuit += RZ(2.0 * gamma)(0)
        target_circuit += RZ(2.0 * gamma)(1)
        target_circuit += RY(-thetas[0])(0)
        target_circuit += RY(-thetas[1])(1)
        target_circuit += RZ(-2.0 * beta)(0)
        target_circuit += RZ(-2.0 * beta)(1)
        target_circuit += RY(thetas[0])(0)
        target_circuit += RY(thetas[1])(1)

    return target_circuit.bind(symbols_map).to_unitary()


class TestWarmStartQAOAAnsatz(AnsatzTests):
    @pytest.fixture
    def thetas(self):
        return np.array([0.5, 0.5])

    @pytest.fixture(params=[1, 2, 3])
    def number_of_layers(self, request):
        return request.param

    @pytest.fixture
    def ansatz(self, thetas, number_of_layers):
        cost_hamiltonian = QubitOperator((0, "Z")) + QubitOperator((1, "Z"))
        return WarmStartQAOAAnsatz(
            number_of_layers=number_of_layers,
            cost_hamiltonian=cost_hamiltonian,
            thetas=thetas,
        )

    def test_set_cost_hamiltonian(self, ansatz):
        # Given
        new_cost_hamiltonian = QubitOperator((0, "Z")) - QubitOperator((1, "Z"))

        # When
        ansatz.cost_hamiltonian = new_cost_hamiltonian

        # Then
        assert ansatz._cost_hamiltonian == new_cost_hamiltonian

    def test_set_cost_hamiltonian_invalidates_circuit(self, ansatz):
        # Given
        new_cost_hamiltonian = QubitOperator((0, "Z")) - QubitOperator((1, "Z"))

        # When
        ansatz.cost_hamiltonian = new_cost_hamiltonian

        # Then
        assert ansatz._parametrized_circuit is None

    def test_set_thetas(self, ansatz):
        # Given
        new_thetas = np.array([0.3, 0.3])

        # When
        ansatz.thetas = new_thetas

        # Then
        ansatz._thetas == new_thetas

    def test_set_thetas_invalidates_circuit(self, ansatz):
        # Given
        new_thetas = np.array([0.3, 0.3])

        # When
        ansatz.thetas = new_thetas

        # Then
        assert ansatz._parametrized_circuit is None

    def test_get_number_of_qubits(self, ansatz):
        # Given
        new_cost_hamiltonian = (
            QubitOperator((0, "Z")) + QubitOperator((1, "Z")) + QubitOperator((2, "Z"))
        )
        target_number_of_qubits = 3

        # When
        ansatz.cost_hamiltonian = new_cost_hamiltonian

        # Then
        assert ansatz.number_of_qubits == target_number_of_qubits

    def test_get_number_of_qubits_with_ising_hamiltonian(self, ansatz):
        # Given
        new_cost_hamiltonian = (
            QubitOperator((0, "Z")) + QubitOperator((1, "Z")) + QubitOperator((2, "Z"))
        )
        new_cost_hamiltonian = change_operator_type(new_cost_hamiltonian, IsingOperator)
        target_number_of_qubits = 3

        # When
        ansatz.cost_hamiltonian = new_cost_hamiltonian

        # Then
        assert ansatz.number_of_qubits == target_number_of_qubits

    def test_get_parametrizable_circuit(self, ansatz, number_of_layers):
        # Then
        betas = create_betas(number_of_layers)
        gammas = create_gammas(number_of_layers)
        target_params = []
        for i in range(number_of_layers):
            target_params.append(gammas[i])
            target_params.append(betas[i])

        assert ansatz.parametrized_circuit.free_symbols == target_params

    def test_generate_circuit(self, ansatz, number_of_layers, thetas):
        # When
        symbols_map = create_symbols_map(number_of_layers)
        parametrized_circuit = ansatz._generate_circuit()
        evaluated_circuit = parametrized_circuit.bind(symbols_map)
        final_unitary = evaluated_circuit.to_unitary()
        target_unitary = create_target_unitary(thetas, number_of_layers)
        # Then
        assert compare_unitary(final_unitary, target_unitary, tol=1e-10)

    def test_generate_circuit_with_ising_operator(
        self, ansatz, number_of_layers, thetas
    ):
        # When
        ansatz.cost_hamiltonian = change_operator_type(
            ansatz.cost_hamiltonian, IsingOperator
        )

        parametrized_circuit = ansatz._generate_circuit()
        symbols_map = create_symbols_map(number_of_layers)
        target_unitary = create_target_unitary(thetas, number_of_layers)
        evaluated_circuit = parametrized_circuit.bind(symbols_map)
        final_unitary = evaluated_circuit.to_unitary()

        # Then
        assert compare_unitary(final_unitary, target_unitary, tol=1e-10)


def test_convert_relaxed_solution_to_angles():
    relaxed_solution = np.array([0, 0.5, 1])
    epsilon = 0.1
    target_converted_solution = np.array(
        [
            2 * np.arcsin(np.sqrt(epsilon)),
            2 * np.arcsin(np.sqrt(0.5)),
            2 * np.arcsin(np.sqrt(1 - epsilon)),
        ]
    )
    converted_solution = convert_relaxed_solution_to_angles(
        relaxed_solution, epsilon=epsilon
    )
    assert np.allclose(converted_solution, target_converted_solution)


def test_convert_relaxed_solution_to_angles_throws_exception_for_invalid_parameters():
    relaxed_solution = np.array([-1, 2, 1])

    with pytest.raises(ValueError):
        convert_relaxed_solution_to_angles(relaxed_solution, epsilon=0.1)

==> orquestra-vqa-0.7.0.tar.gz: orquestra-vqa-0.7.0/tests/orquestra/vqa/ansatz/qaoa_warm_start_test.py <==
################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
import numpy as np
import pytest
import sympy
from orquestra.quantum.circuits import RY, RZ, Circuit
from orquestra.quantum.operators import PauliSum, PauliTerm
from orquestra.quantum.utils import compare_unitary

from orquestra.vqa.ansatz.qaoa_warm_start import (
    WarmStartQAOAAnsatz,
    convert_relaxed_solution_to_angles,
)
from orquestra.vqa.api.ansatz_test import AnsatzTests


def create_betas(number_of_layers):
    return sympy.symbols(f"beta_:{number_of_layers}")


def create_gammas(number_of_layers):
    return sympy.symbols(f"gamma_:{number_of_layers}")


def create_symbols_map(number_of_layers):
    betas = create_betas(number_of_layers)
    gammas = create_gammas(number_of_layers)
    symbols_map = []
    beta_value = 0.5
    gamma_value = 0.7
    for beta, gamma in zip(betas, gammas):
        symbols_map.append([beta, beta_value])
        symbols_map.append([gamma, gamma_value])
    return symbols_map


def create_target_unitary(thetas, number_of_layers):
    target_circuit = Circuit()
    target_circuit += RY(thetas[0])(0)
    target_circuit += RY(thetas[1])(1)
    betas = create_betas(number_of_layers)
    gammas = create_gammas(number_of_layers)
    symbols_map = create_symbols_map(number_of_layers)
    for layer_id in range(number_of_layers):
        beta = betas[layer_id]
        gamma = gammas[layer_id]
        target_circuit += RZ(2.0 * gamma)(0)
        target_circuit += RZ(2.0 * gamma)(1)
        target_circuit += RY(-thetas[0])(0)
        target_circuit += RY(-thetas[1])(1)
        target_circuit += RZ(-2.0 * beta)(0)
        target_circuit += RZ(-2.0 * beta)(1)
        target_circuit += RY(thetas[0])(0)
        target_circuit += RY(thetas[1])(1)

    return target_circuit.bind(symbols_map).to_unitary()


class TestWarmStartQAOAAnsatz(AnsatzTests):
    @pytest.fixture
    def thetas(self):
        return np.array([0.5, 0.5])

    @pytest.fixture(params=[1, 2, 3])
    def number_of_layers(self, request):
        return request.param

    @pytest.fixture
    def ansatz(self, thetas, number_of_layers):
        cost_hamiltonian = PauliSum("Z0+Z1")
        return WarmStartQAOAAnsatz(
            number_of_layers=number_of_layers,
            cost_hamiltonian=cost_hamiltonian,
            thetas=thetas,
        )

    def test_set_cost_hamiltonian(self, ansatz):
        # Given
        new_cost_hamiltonian = PauliSum("Z0 + -1*Z1")

        # When
        ansatz.cost_hamiltonian = new_cost_hamiltonian

        # Then
        assert ansatz._cost_hamiltonian == new_cost_hamiltonian

    def test_set_cost_hamiltonian_invalidates_circuit(self, ansatz):
        # Given
        new_cost_hamiltonian = PauliSum("Z0 + -1*Z1")

        # When
        ansatz.cost_hamiltonian = new_cost_hamiltonian

        # Then
        assert ansatz._parametrized_circuit is None

    def test_set_thetas(self, ansatz):
        # Given
        new_thetas = np.array([0.3, 0.3])

        # When
        ansatz.thetas = new_thetas

        # Then
        ansatz._thetas == new_thetas

    def test_set_thetas_invalidates_circuit(self, ansatz):
        # Given
        new_thetas = np.array([0.3, 0.3])

        # When
        ansatz.thetas = new_thetas

        # Then
        assert ansatz._parametrized_circuit is None

    def test_get_number_of_qubits(self, ansatz):
        # Given
        new_cost_hamiltonian = PauliSum("Z0+Z1+Z2")
        target_number_of_qubits = 3

        # When
        ansatz.cost_hamiltonian = new_cost_hamiltonian

        # Then
        assert ansatz.number_of_qubits == target_number_of_qubits

    def test_get_number_of_qubits_with_pauli_term(self, ansatz):
        # Given
        new_cost_hamiltonian = PauliTerm("Z0*Z1*Z2")
        target_number_of_qubits = 3

        # When
        ansatz.cost_hamiltonian = new_cost_hamiltonian

        # Then
        assert ansatz.number_of_qubits == target_number_of_qubits

    def test_get_parametrizable_circuit(self, ansatz, number_of_layers):
        # Then
        betas = create_betas(number_of_layers)
        gammas = create_gammas(number_of_layers)
        target_params = []
        for i in range(number_of_layers):
            target_params.append(gammas[i])
            target_params.append(betas[i])

        assert ansatz.parametrized_circuit.free_symbols == target_params

    def test_generate_circuit(self, ansatz, number_of_layers, thetas):
        # When
        symbols_map = create_symbols_map(number_of_layers)
        parametrized_circuit = ansatz._generate_circuit()
        evaluated_circuit = parametrized_circuit.bind(symbols_map)
        final_unitary = evaluated_circuit.to_unitary()
        target_unitary = create_target_unitary(thetas, number_of_layers)
        # Then
        assert compare_unitary(final_unitary, target_unitary, tol=1e-10)

    def test_generate_circuit_with_pauli_term(self, ansatz, number_of_layers):
        # Given
        thetas = np.array([0.5])

        def create_1_qubit_target_unitary(thetas, number_of_layers):
            target_circuit = Circuit()
            target_circuit += RY(thetas[0])(0)
            betas = create_betas(number_of_layers)
            gammas = create_gammas(number_of_layers)
            symbols_map = create_symbols_map(number_of_layers)
            for layer_id in range(number_of_layers):
                beta = betas[layer_id]
                gamma = gammas[layer_id]
                target_circuit += RZ(2.0 * gamma)(0)
                target_circuit += RY(-thetas[0])(0)
                target_circuit += RZ(-2.0 * beta)(0)
                target_circuit += RY(thetas[0])(0)

            return target_circuit.bind(symbols_map).to_unitary()

        # When
        ansatz = WarmStartQAOAAnsatz(number_of_layers, PauliTerm("Z0"), thetas)

        parametrized_circuit = ansatz._generate_circuit()
        symbols_map = create_symbols_map(number_of_layers)
        target_unitary = create_1_qubit_target_unitary(thetas, number_of_layers)
        evaluated_circuit = parametrized_circuit.bind(symbols_map)
        final_unitary = evaluated_circuit.to_unitary()

        # Then
        assert compare_unitary(final_unitary, target_unitary, tol=1e-10)


def test_convert_relaxed_solution_to_angles():
    relaxed_solution = np.array([0, 0.5, 1])
    epsilon = 0.1
    target_converted_solution = np.array(
        [
            2 * np.arcsin(np.sqrt(epsilon)),
            2 * np.arcsin(np.sqrt(0.5)),
            2 * np.arcsin(np.sqrt(1 - epsilon)),
        ]
    )
    converted_solution = convert_relaxed_solution_to_angles(
        relaxed_solution, epsilon=epsilon
    )
    assert np.allclose(converted_solution, target_converted_solution)


def test_convert_relaxed_solution_to_angles_throws_exception_for_invalid_parameters():
    relaxed_solution = np.array([-1, 2, 1])

    with pytest.raises(ValueError):
        convert_relaxed_solution_to_angles(relaxed_solution, epsilon=0.1)
